
### Source codes accompanying the paper ``Illumination depth'' by Stanislav Nagy and Jiri Dvorak,
### Journal of Computational and Graphical Statistics.

### This file provides an example of the classification procedures for bivariate normal distributions,
### given in Section 5.3.1 of the paper.

### Version 7.2.2020


### Required external packages ###
##################################

library(splancs)
library(mvtnorm)
library(depth)
library(TukeyRegion)

# Auxiliary variables, delta specifies the upper level set of
# the halfspace depth used for illumination
delta <- 1 - pnorm(sqrt(2*log(2)))
q.delta <- qnorm(1-delta)


### Auxiliary functions for the illumination part ###
#####################################################

Ill.aff = function(x,pol){
  # compute standardized illumination of point x w.r.t. the convex hull of given set of points
  Z = rbind(pol,x)
  return(Area(Z)/Area(pol))
}

Area = function(X){
  # compute area of the convex hull of a set of points in two dimensions
  chull(X[,1],X[,2])->i
  return(areapl(cbind(X[i,1],X[i,2])))
}

g.fun <- function(y){
  # function g_d introduced in Section 2.1 of the paper, explicit formula for dimension d=2
  1 + (2/pi)*((y/2)*(1 - 1/y^2)^(3/2) - (1/2*acos(1/y) - 1/(2*y)*sqrt(1-1/(y^2))))
}

arguments <- c(seq(from=1, to=2, by=0.001),seq(from=2.01, to=6, by=0.01),seq(from=6.1, to=20, by=0.1))
g.vals <- g.fun(arguments)
arg.last <- arguments[length(arguments)]
g.last <- g.vals[length(g.vals)]
slope.last <- (arg.last-arguments[length(arguments)-1])/(g.last-g.vals[length(g.vals)-1])

g.inv.fun <- function(y){
  # inverse function to g_2
  if (y==1){return(1)}
  if (y >= g.last){return(arg.last+(y-g.last)*slope.last)}
  else
  {
    kde <- sum(g.vals < y)
    return(arguments[kde] + (y-g.vals[kde])*(arguments[kde+1]-arguments[kde])/(g.vals[kde+1]-g.vals[kde]))
  }
}



### Location difference, scale difference, no contamination, classify all points ###
####################################################################################

set.seed(2019)

m1 = c(0,0)
m2 = c(4,4)
rho = 0
S1 = matrix(c(1,rho,rho,1),ncol=2)
S2 = matrix(c(4,rho,rho,4),ncol=2)
n = 500
n.test = 1000

misclass.rate.ls <- rep(NA, times=100)         # illumination
misclass.rate.qda.ls <- rep(NA, times=100)     # standard QDA

k1 <- rep(0,100)
k2 <- rep(0,100)

system.time({
  for (opak in 1:100){
    
    if ((opak %% 10) == 0){cat(opak)} else {cat(".")}
    
    X1 = rmvnorm(n,m1,S1)
    X2 = rmvnorm(n,m2,S2)
    
    DX1 = rep(NA,n)
    DX2 = rep(NA,n)
    for(i in 1:n){
      DX1[i] = depth(X1[i,],X1)
      DX2[i] = depth(X2[i,],X2)
    }
    
    v1 <- Area(X1[DX1>delta,])
    v2 <- Area(X2[DX2>delta,])
    
    Y1 <- rmvnorm(n=n.test,m1,S1)
    Y2 <- rmvnorm(n=n.test,m2,S2)
    
    DY1X1 = rep(NA,n.test)
    DY2X1 = rep(NA,n.test)
    DY1X2 = rep(NA,n.test)
    DY2X2 = rep(NA,n.test)
    for(i in 1:n.test){
      DY1X1[i] = depth(Y1[i,],X1)
      DY2X1[i] = depth(Y2[i,],X1)
      DY1X2[i] = depth(Y1[i,],X2)
      DY2X2[i] = depth(Y2[i,],X2)
    }
    
    
    ### Illumination-based classifier ###
    TR1 <- TukeyRegion(data=X1, depth=delta*n, retVertices = TRUE)$vertices
    ch1 <- chull(TR1[,1],TR1[,2])
    inner1 <- cbind(TR1[ch1,1],TR1[ch1,2])
    
    TR2 <- TukeyRegion(data=X2, depth=delta*n, retVertices = TRUE)$vertices
    ch2 <- chull(TR2[,1],TR2[,2])
    inner2 <- cbind(TR2[ch2,1],TR2[ch2,2])
    
    #######################################################
    MY1X1 <- rep(NA,times=n.test)
    which.outside <- (DY1X1 < delta)
    Y1.outside <- Y1[which.outside,]
    val <- rep(NA, times=sum(which.outside))
    for (i in 1:sum(which.outside)){
      pom <- Ill.aff(x=Y1.outside[i,],pol=inner1)
      val[i] <- g.inv.fun(pom)*q.delta
    }
    MY1X1[DY1X1 >= delta] <- qnorm(1-DY1X1[DY1X1>=delta])
    MY1X1[DY1X1 < delta]  <- val
    
    #######################################################
    MY2X1 <- rep(NA,times=n.test) 
    which.outside <- (DY2X1 < delta)
    Y2.outside <- Y2[which.outside,]
    val <- rep(NA, times=sum(which.outside))
    for (i in 1:sum(which.outside)){
      pom <- Ill.aff(x=Y2.outside[i,],pol=inner1)
      val[i] <- g.inv.fun(pom)*q.delta
    }
    MY2X1[DY2X1 >= delta] <- qnorm(1-DY2X1[DY2X1>=delta])
    MY2X1[DY2X1 < delta]  <- val
    
    #######################################################
    MY1X2 <- rep(NA,times=n.test)
    which.outside <- (DY1X2 < delta)
    Y1.outside <- Y1[which.outside,]
    val <- rep(NA, times=sum(which.outside))
    for (i in 1:sum(which.outside)){
      pom <- Ill.aff(x=Y1.outside[i,],pol=inner2)
      val[i] <- g.inv.fun(pom)*q.delta
    }
    MY1X2[DY1X2 >= delta] <- qnorm(1-DY1X2[DY1X2>=delta])
    MY1X2[DY1X2 < delta]  <- val
    
    #######################################################
    MY2X2 <- rep(NA,times=n.test) 
    which.outside <- (DY2X2 < delta)
    Y2.outside <- Y2[which.outside,]
    val <- rep(NA, times=sum(which.outside))
    for (i in 1:sum(which.outside)){
      pom <- Ill.aff(x=Y2.outside[i,],pol=inner2)
      val[i] <- g.inv.fun(pom)*q.delta
    }
    MY2X2[DY2X2 >= delta] <- qnorm(1-DY2X2[DY2X2>=delta])
    MY2X2[DY2X2 < delta]  <- val
    
    vals11 <- -2*log(v1) - MY1X1^2
    vals21 <- -2*log(v1) - MY2X1^2
    vals12 <- -2*log(v2) - MY1X2^2
    vals22 <- -2*log(v2) - MY2X2^2
    
    misclass.rate.ls[opak] <- mean(c(vals11 <= vals12, vals21 > vals22))
    
    
    ### Classical QDA classifier ###
    class.train <- c(rep(1,n),rep(2,n))
    class.test <- c(rep(1,n.test),rep(2,n.test))
    data.train <- data.frame(class.train,rbind(X1,X2))
    z.qda <- qda(class.train ~ ., data.train, prior = c(1/2,1/2))
    pred.qda <- predict(z.qda, data.frame(rbind(Y1,Y2)))$class
    misclass.rate.qda.ls[opak] <- mean(pred.qda != class.test)
  }
})


op = par(cex=1.5, mar=c(2,2,2,0.5))
boxplot(misclass.rate.ls,misclass.rate.qda.ls, lwd=2, ylim=c(0,0.06), names=c("Illumination", "QDA"), main="All points", cex=1)
# Bayes error rate (calculated elsewhere): 0.0241552
abline(a=0.0241552, b=0, lty=2, lwd=2, col="black")

summary(misclass.rate.ls)
summary(misclass.rate.qda.ls)

c(mean(misclass.rate.ls),sd(misclass.rate.ls))
c(mean(misclass.rate.qda.ls),sd(misclass.rate.qda.ls))



### Location difference, scale difference, no contamination, classify only outsiders ###
########################################################################################

set.seed(2019)

m1 = c(0,0)
m2 = c(4,4)
rho = 0
S1 = matrix(c(1,rho,rho,1),ncol=2)
S2 = matrix(c(4,rho,rho,4),ncol=2)
n = 500
n.test = 2500

misclass.rate.ls.outer <- rep(NA, times=100)         # illumination
misclass.rate.qda.ls.outer <- rep(NA, times=100)     # standard QDA

k1 <- rep(0,100)
k2 <- rep(0,100)

system.time({
  for (opak in 1:100){
    
    if ((opak %% 10) == 0){cat(opak)} else {cat(".")}
    
    X1 = rmvnorm(n,m1,S1)
    X2 = rmvnorm(n,m2,S2)
    
    DX1 = rep(NA,n)
    DX2 = rep(NA,n)
    for(i in 1:n){
      DX1[i] = depth(X1[i,],X1)
      DX2[i] = depth(X2[i,],X2)
    }
    
    v1 <- Area(X1[DX1>delta,])
    v2 <- Area(X2[DX2>delta,])
    
    Y1 <- rmvnorm(n=n.test,m1,S1)
    Y2 <- rmvnorm(n=n.test,m2,S2)
    
    DY1X1 = rep(NA,n.test)
    DY2X1 = rep(NA,n.test)
    DY1X2 = rep(NA,n.test)
    DY2X2 = rep(NA,n.test)
    for(i in 1:n.test){
      DY1X1[i] = depth(Y1[i,],X1)
      DY2X1[i] = depth(Y2[i,],X1)
      DY1X2[i] = depth(Y1[i,],X2)
      DY2X2[i] = depth(Y2[i,],X2)
    }
    
    
    ### Illumination-based classifier ###
    TR1 <- TukeyRegion(data=X1, depth=delta*n, retVertices = TRUE)$vertices
    ch1 <- chull(TR1[,1],TR1[,2])
    inner1 <- cbind(TR1[ch1,1],TR1[ch1,2])
    
    TR2 <- TukeyRegion(data=X2, depth=delta*n, retVertices = TRUE)$vertices
    ch2 <- chull(TR2[,1],TR2[,2])
    inner2 <- cbind(TR2[ch2,1],TR2[ch2,2])
    
    #######################################################
    which.outside <- (DY1X1 == 0) & (DY1X2 == 0)
    Y1.outside <- Y1[which.outside,]
    val <- rep(NA, times=sum(which.outside))
    MY1X1 <- rep(NA,times=sum(which.outside))
    for (i in 1:sum(which.outside)){
      pom <- Ill.aff(x=Y1.outside[i,],pol=inner1)
      val[i] <- g.inv.fun(pom)*q.delta
    }
    MY1X1  <- val
    
    #######################################################
    which.outside <- (DY2X1 == 0) & (DY2X2 == 0)
    Y2.outside <- Y2[which.outside,]
    val <- rep(NA, times=sum(which.outside))
    MY2X1 <- rep(NA,times=sum(which.outside)) 
    for (i in 1:sum(which.outside)){
      pom <- Ill.aff(x=Y2.outside[i,],pol=inner1)
      val[i] <- g.inv.fun(pom)*q.delta
    }
    MY2X1  <- val
    
    #######################################################
    which.outside <- (DY1X1 == 0) & (DY1X2 == 0)
    Y1.outside <- Y1[which.outside,]
    val <- rep(NA, times=sum(which.outside))
    MY1X2 <- rep(NA,times=sum(which.outside))
    for (i in 1:sum(which.outside)){
      pom <- Ill.aff(x=Y1.outside[i,],pol=inner2)
      val[i] <- g.inv.fun(pom)*q.delta
    }
    MY1X2  <- val
    
    #######################################################
    which.outside <- (DY2X1 == 0) & (DY2X2 == 0)
    Y2.outside <- Y2[which.outside,]
    val <- rep(NA, times=sum(which.outside))
    MY2X2 <- rep(NA,times=sum(which.outside)) 
    for (i in 1:sum(which.outside)){
      pom <- Ill.aff(x=Y2.outside[i,],pol=inner2)
      val[i] <- g.inv.fun(pom)*q.delta
    }
    MY2X2  <- val
    
    vals11 <- -2*log(v1) - MY1X1^2
    vals21 <- -2*log(v1) - MY2X1^2
    vals12 <- -2*log(v2) - MY1X2^2
    vals22 <- -2*log(v2) - MY2X2^2
    
    misclass.rate.ls.outer[opak] <- mean(c(vals11 <= vals12, vals21 > vals22))
    
    
    ### Classical QDA classifier ###
    which.outside1 <- (DY1X1 == 0) & (DY1X2 == 0)
    which.outside2 <- (DY2X1 == 0) & (DY2X2 == 0)
    class.train <- c(rep(1,n),rep(2,n))
    class.test <- c(rep(1,sum(which.outside1)),rep(2,sum(which.outside2)))
    data.train <- data.frame(class.train,rbind(X1,X2))
    z.qda <- qda(class.train ~ ., data.train, prior = c(1/2,1/2))
    pred.qda <- predict(z.qda, data.frame(rbind(Y1[which.outside1,],Y2[which.outside2,])))$class
    misclass.rate.qda.ls.outer[opak] <- mean(pred.qda != class.test)
  }
})


op = par(cex=1.5, mar=c(2,2,2,0.5))
boxplot(misclass.rate.ls.outer,misclass.rate.qda.ls.outer, lwd=2, ylim=c(0,0.15), names=c("Illumination", "QDA"), main="Outsiders", cex=1)

c(mean(misclass.rate.ls.outer),sd(misclass.rate.ls.outer))
c(mean(misclass.rate.qda.ls.outer),sd(misclass.rate.qda.ls.outer))