
### Source codes accompanying the paper ``Illumination depth'' by Stanislav Nagy and Jiri Dvorak,
### Journal of Computational and Graphical Statistics.

### This file provides an example of the procedure for estimating extreme central regions
### given in Section 5.2 of the paper.

### Version 7.2.2020


### Required external packages ###
##################################

library(LaplacesDemon)
library(splancs)
library(TukeyRegion)
library(plotrix)


### Auxiliary functions ###
###########################

Ill.aff = function(x,pol){
  # compute standardized illumination of point x w.r.t. the convex hull of given set of points
  Z = rbind(pol,x)
  return(Area(Z)/Area(pol))
}

Area = function(X){
  # compute area of the convex hull of a set of points in two dimensions
  chull(X[,1],X[,2])->i
  return(areapl(cbind(X[i,1],X[i,2])))
}

g.fun <- function(y){ 
  # function g_d introduced in Section 2.1 of the paper, explicit formula for dimension d=2
  1 + (2/pi)*((y/2)*(1 - 1/y^2)^(3/2) - (1/2*acos(1/y) - 1/(2*y)*sqrt(1-1/(y^2))))
}


### Estimation of the extreme central regions ###
#################################################

# generate data from 2-dimensional Cauchy distribution
set.seed(2019)
n <- 500 # sample size
k <- 75  # tuning parameter, see Section 5.2 of the paper
u <- 1/n # required depth level for the extreme region
data <- rmvc(n, c(0,0), diag(2))

# determine the set S=P_{n,k/n}
TR <- TukeyRegion(data=data, depth=(k/n)*n, retVertices = TRUE)$vertices
ch <- chull(TR[,1],TR[,2])
inner <- cbind(TR[ch,1],TR[ch,2])

# estimete the tail index alpha and determine the scaling factor c
s <- sqrt(rowSums(data^2))
s <- sort(s)
gamma <- sum(log(s[(n-k+1):n]/s[n-k]))/k
alpha.hat <- 1/gamma
coeff <- (k/(u*n))^(1/alpha.hat)

# determine illumination of points on the given grid
nr <- 200
rs <- seq(from=30, to=200, length.out=nr)
nc <- 500
thetas <- seq(from=0, to=2*pi*(1-1/nc), length.out=nc)
mat <- matrix(NA, nrow=nr, ncol=nc) 
for (i in 1:nr){
  for (j in 1:nc){
    mat[i,j] <- Ill.aff(c(rs[i]*cos(thetas[j]),rs[i]*sin(thetas[j])),pol=inner)
  }
}

# find the set E_I
tresh <- g.fun(coeff)
levset <- matrix(NA, nrow=nc, ncol=2)
for (j in 1:nc){
  aux <- sum(mat[,j]<tresh)+1
  levset[j,1] <- rs[aux]*cos(thetas[j])
  levset[j,2] <- rs[aux]*sin(thetas[j])
}


# plot the estimated extreme central region: 
# blue polygon ... estimate based on the refined depth
# orange polygon ... estimate based on the illumination
# black circle ..... boundary of the true (theoretical) extreme central region

op = par(cex=1.5, mar=c(2,2,0.5,0.5))
plot(data, asp=1, xlim=c(-200,700), ylim=c(-300,300), cex=2/3)
# a data point with coordinates (-1676.647, -1691.202) is not plotted!
polygon(coeff*inner, border="blue", lwd=2, col=adjustcolor("blue", alpha.f = 0.25))
polygon(levset[,1],levset[,2], border="orange", lwd=2, col=adjustcolor("orange", alpha.f = 0.25))
draw.circle(x=0,y=0,radius=qcauchy(1-u), lwd=2, lty=1)