
### Source codes accompanying the paper ``Illumination depth'' by Stanislav Nagy and Jiri Dvorak,
### Journal of Computational and Graphical Statistics.

### This file provides an example of the tie-breaking procedure given
### in Section 5.1 of the paper.

### Version 7.2.2020


### Required external packages ###
##################################

library(depth)
library(mvtnorm)
library(Matrix)
library(geometry)


### Auxiliary functions ###
###########################

Volume = function(X){
  # compute volume of the convex hull of X
  if((is.null(nrow(X)))||(nrow(X)==0)) return(0)
  return(convhulln(X, options="FA")$vol)
}


### Main function ###
#####################

Ill.nD.tiebreaking = function(x,X,D=NULL,sD=NULL, frac=0.5){
  # compute illumination for tie-breaking purposes
  
  # x ...... d-dimensional vector, point for which the illumination is computed
  # X ...... n x d matrix containing the n observations in dimension d
  # D ...... vector of sample halfspace depths for the observations
  # sD ..... sorted version of D
  # frac ... which fraction of points deeper than x should be included in the central region

  n = nrow(X)
  dimension = ncol(X)
  
  # if sample halfspace depth values are not provided as argument, compute them
  if(is.null(D)){
    D = rep(NA,n)
    for(i in 1:n) D[i] = depth(X[i,],X[-i,], approx=TRUE)
  }
  
  # if sorted values of the sample halfspace depth are not provided, compute them
  if(is.null(sD)){sD = sort(D)}
  
  # determine the central region as the halfspace depth upper level set 
  # containing "frac" of the points with depth at least the depth of x
  aux <- depth(x,X, approx=TRUE)
  aux2 <- floor(sum(D >= aux)*frac)
  delta <- sD[aux2]
  Y <- matrix(X[D>=delta,],ncol=dimension)
  Z <- rbind(Y,x)
  
  # determine the illumination
  if (nrow(Z) < (dimension+1)){ill = 0} else {ill <- Volume(Z)}
  return(ill)
}


### Example of tie-breaking procedure ###
#########################################

# generate data from 5-dimensional normal distribution
set.seed(12345)
dimension <- 5
n <- 1000
m <- rep(0, times=dimension)
S <- diag(nrow=dimension)
X <- rmvnorm(n,m,S)


# correct ranking based on the values of the probability density function
ranking.correct <- rank(sqrt(rowSums(X^2)))


# ranking based on halfspace depth
hD <- rep(NA,nrow(X))
for(i in 1:nrow(X)) hD[i] <- depth(X[i,],X, approx=TRUE)
ranking.hD <- rank(-hD, ties.method = "average")


# ranking based on halfspace depth, ties resolved based on illumination
ill <- rep(NA,nrow(X))
sD <- sort(hD,decreasing=TRUE)
for(i in 1:nrow(X)) ill[i] <- Ill.nD.tiebreaking(x=X[i,],X=X,D=hD,sD=sD,frac=1/2)
ranking.ill <- invPerm(order(-hD,ill))


# ranking performance on complete data
cor(ranking.correct,ranking.hD,method="spearman")
cor(ranking.correct,ranking.ill,method="spearman")

plot(ranking.correct,ranking.hD,ylim=c(0,n),xlab="ranking.correct",ylab="ranking.hD",main="All observations")
plot(ranking.correct,ranking.ill,ylim=c(0,n),xlab="ranking.correct",ylab="ranking.ill",main="All observations")


# ranking performance on data points with halfspace depth <= 0.01
aux <- (hD<=0.01)
cor(ranking.correct[aux],ranking.hD[aux],method="spearman")
cor(ranking.correct[aux],ranking.ill[aux],method="spearman")

plot(ranking.correct[aux],ranking.hD[aux],ylim=c(600,n),xlab="ranking.correct",ylab="ranking.hD",main="Observations with hD <= 0.01")
plot(ranking.correct[aux],ranking.ill[aux],ylim=c(600,n),xlab="ranking.correct",ylab="ranking.ill",main="Observations with hD <= 0.01")